/*  CLIPPER PROGRAMMING GUIDE:  PROGRAM 12-12  Memo Pack Buffering
 *
 *                 pp 560 - 563
 *  buff.c
 */
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdio.h>
#include <io.h>
#include "dbf.h"
#include "extend.h"

#define O_RAW  O_BINARY       /* QC2 - BC compatibility */

/*  BUFFER Control Structure */
typedef struct
{
    unsigned dbt_rec_no;
    char *buff;
    unsigned size_buff;
    unsigned num_dbt_recs;
} BUFFER;

/* Here are two instances of it! */
BUFFER read_buff;
BUFFER write_buff;

/* Required Constants */
#define MEMO_SIZE       512
#define READ_RECS         8
#define WRITE_RECS       32
#define READ_BUFF_SIZE  READ_RECS*MEMO_SIZE
#define WRITE_BUFF_SIZE WRITE_RECS*MEMO_SIZE

/*   BUFF_INIT()
 *                   Initialize buffers, allocate buffers
 */

Boolean buff_init(void)
{
    read_buff.size_buff    = READ_BUFF_SIZE;
    read_buff.dbt_rec_no   = 1;
    read_buff.num_dbt_recs = 0;

    read_buff.buff = (char *)malloc(READ_BUFF_SIZE);
    if (!read_buff.buff)
    {
        D(printf("\n*** BUFF_INIT:  READ BUFFER FAILURE ***"); )
        return (FALSE);
    }

    write_buff.size_buff    = WRITE_BUFF_SIZE;
    write_buff.buff = (char *)malloc(WRITE_BUFF_SIZE);
    if (!write_buff.buff)
    {
        D(printf("\n*** BUFF_INIT:  WRITE BUFFER FAILURE ***"); )
        return (FALSE);
    }
    write_buff.dbt_rec_no   = 1;
    write_buff.num_dbt_recs = 0;
                               D(printf("\nREAD/WRITE BUFFers INITialized."); )
    return(TRUE);
}
/* -------------------------------------------------------------- */
/*  BUFF_END()
 *              Finished with buffers, release memory & flush write buffer
 */
void buff_end(int handle)
{
    lseek ( handle,  (long)write_buff.dbt_rec_no * MEMO_SIZE,   SEEK_SET);
    write ( handle,  write_buff.buff,  write_buff.num_dbt_recs*MEMO_SIZE);
    free  (write_buff.buff);
    free  (read_buff.buff );
                               D(printf("\nREAD/WRITE BUFFers ENDed."); )

}
/* --------------------------------------------------------------- */
/*  GET_DBT_REC
 *              Return pointer to required DBT Record.
 *              If record in buffer already, return it from there,
 *              saving disk reads.  If not, read buffer full
 *              starting from this record (readahead).
 */
char *get_dbt_rec(int handle, unsigned rec_no)
{
    /* If requested record not in buffer, Read it! */
    if (rec_no < read_buff.dbt_rec_no ||
        rec_no > read_buff.dbt_rec_no + read_buff.num_dbt_recs - 1)
    {
        /* Seek to start of record */
        lseek (handle, (long) rec_no * MEMO_SIZE, SEEK_SET);
        /* Read Buffer-full (Readahead) */
        read_buff.num_dbt_recs =
            (read(handle, read_buff.buff, read_buff.size_buff) +
            MEMO_SIZE - 1) / MEMO_SIZE;
        read_buff.dbt_rec_no = rec_no;
    }
    /* return pointer to it! */
    return ((char *)&read_buff.buff[
                    (rec_no - read_buff.dbt_rec_no) * MEMO_SIZE]);
}
/* --------------------------------------------------------------- */
/***
 *   PUT_DBT_REC()
 *
 *   Write Specified record to buffer.  If buffer is full, perform
 *     actual disk write to flush it first.
 */

void put_dbt_rec(int handle, unsigned rec_no, char *rec)
{
    /* Flush buffer if full */
    if (write_buff.num_dbt_recs == (write_buff.size_buff / MEMO_SIZE) )
    {
        lseek (handle, (long)write_buff.dbt_rec_no*MEMO_SIZE,SEEK_SET);
        write(handle, write_buff.buff, write_buff.size_buff);
        write_buff.dbt_rec_no = rec_no;
        write_buff.num_dbt_recs = 0;
    }

    /* Place record in buffer */
    memcpy(&write_buff.buff[(rec_no - write_buff.dbt_rec_no)*MEMO_SIZE],
           rec, MEMO_SIZE);

    /* Increment count */
    write_buff.num_dbt_recs++;

}
/* --------------------------------------------------------------- */
/* EOF:  BUFF.C  */
