/*************************************/
/* MEMOPAK1.C
 *               C code to do the actual MEMO PACK
 *                    BUFFERED version (Uses BUFF.C)
 *                    Prog 12-13 in CLIPPER USER'S GUIDE
 *                    p 563, ff
 */
/*  1.  Get parameters from CLIPPER and save them in
        local variables and arrays.                    */
/*  2.  Open MEMO and DBF files.                       */
/*  3.  For each MEMO record                           */
/*      3.1  Call Move_memo to move it                 */
/*           3.1.1  Seek to start of memo field record */
/*           3.1.2  For each physical record occupied  */
/*                  3.1.2.1  Copy it to new location   */
/*      3.2  Save its new location                     */
/*  4.  Shrink file by writing 0 bytes                 */
/*  5.  Update DBT file header with new (smaller)      */
/*      number of records                              */
/*  6.  Find offset of memo field into FIELD_REC       */
/*  7.  For each memo field record:                    */
/*      7.1  Update DBF record to point to             */
/*           new DBT record.                           */
/*******************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <io.h>
#include <malloc.h>
#include <dos.h>

#include "dbf.h"
#include "extend.h"

#define MEMO_SIZE 512
#define ASCII_EOF 0x1A

#define O_RAW  O_BINARY       /* QC2 - BC compatibility */

/* Defined in this file */
unsigned move_memo( int, unsigned, unsigned);
void update_dbf (int, unsigned, unsigned, unsigned, unsigned, unsigned);

/* Defined in BUFF.C */
Boolean buff_init();
void    buff_end(int);
void    put_dbt_rec(int, unsigned, char *);
char   *get_dbt_rec(int, unsigned);

CLIPPER mem_pak()
{
    unsigned *dbt_recs, *dbf_recs;
    unsigned i, num_memos, field_num, size_memo, next_memo;
    int dbt_handle, dbf_handle;
    char dbf_name[13], dbt_name[13];
    unsigned field_off;
    long num_mem_recs;
    unsigned n_written;

    DBF_HEAD dbf_head;
    FIELD_REC field_rec;

                      D( printf("\n\nC_MEMO_PAK - BUFFERED"); )

/*  1.  Get parameters from CLIPPER and save them in   */
/*      local variables and arrays.                    */
    strcpy(dbf_name, _parc(1));
    strcat(dbf_name, ".DBF");

    strcpy(dbt_name, _parc(1));
    strcat(dbt_name, ".DBT");
                            D(  printf("\n\tDATA: %s",dbf_name); )
                            D(  printf("\n\tTEXT: %s",dbt_name); )
    field_num = _parni(2);
    num_memos = _parni(3);
                            D(  printf("\n\t\tFIELD#  = %d",field_num); )
                            D(  printf("\n\t\t# MEMOS = %d",num_memos); )

    dbf_recs = (unsigned *)malloc(num_memos * sizeof(unsigned));
    dbt_recs = (unsigned *)malloc(num_memos * sizeof(unsigned));

    for ( i = 1; i <= num_memos; i++)
    {
        dbf_recs[i-1] = _parni(4,i);
        dbt_recs[i-1] = _parni(5,i);
    }

/*  2.  Open MEMO and DBF files and initialize BUFFERS.              */
	dbt_handle = open(dbt_name, O_RDWR | O_RAW);

    if (dbt_handle == -1)
    {
         printf("\nDBT OPEN FAILED");
        _retl(FALSE);
        return;
	}

	dbf_handle = open(dbf_name, O_RDWR | O_RAW);
    if (dbf_handle == -1)
    {
        printf("\nDBF OPEN FAILED");
        _retl(FALSE);
        return;
	}
                              D(printf("\nDBT/DBF files OPEN:");)
                              D(printf(" %d / %d.", dbt_handle, dbf_handle);)
    if ( !buff_init() )
    {
        printf("\n*** MEM_PAK:  BUFFER FAILURE ***");
		_retl(FALSE);
        return;
	}

/*  3.  For each MEMO record                           */
	next_memo = 1;
	for (i = 0; i < num_memos; i++)
	{
                              D(printf("\nMove MEMO # %d",i);)
                              D(printf(" to %u.",next_memo); )

/*      3.1  Call Move_memo to move it                 */
/*                from -> to                           */

		size_memo = move_memo (dbt_handle, dbt_recs[i], next_memo);

/*           3.1.1  Seek to start of memo field record */
/*           3.1.2  For each physical record occupied  */
/*                  3.1.2.1  Copy it to new location   */
/*      3.2  Save its new location                     */
		/* where it is now (just moved there ) */

        dbt_recs[i] = next_memo;

        /* where next one will go */
		next_memo  += size_memo ;
	}
    buff_end(dbt_handle);

/*  4.  Shrink file by writing 0 bytes                 */

/*  MICROSOFT vs BORLAND compatibility macro test */
#if (defined (__BORLANDC__)) || (defined (__TURBOC__))
    n_written = (unsigned) _write( dbt_handle, NULL, 0);
#else
	_dos_write(dbt_handle, NULL, 0, &n_written);
#endif

    lseek (dbt_handle, 0L, SEEK_SET);

                                       D(printf("\n File SHRUNK");)

/*  5.  Update DBT file header with new (smaller)      */
/*		number of records (including header)		   */
	num_mem_recs = next_memo;
	write(dbt_handle, (char *)&num_mem_recs, 4);
	close (dbt_handle);

                                       D(printf("\nDBT header updated.");)

    dbf_handle = open(dbf_name, O_RDWR | O_RAW); /* ALREADY OPEN?! */
    read (dbf_handle, (char *)&dbf_head, sizeof(dbf_head));

/*  6.  Find offset of memo field into FIELD_REC       */
/*      N.B.  do this loop only until the              */
/*            ONE AND ONLY MEMO Field occurs           */

                                       D(printf("\nMEMO offset =");)

    field_off = 1;  /* Skip Delete Flag at top of record */
    for (i = 0; i < (field_num - 1); i++)
    {
        read(dbf_handle, (char *)&field_rec, sizeof(FIELD_REC));
        switch (field_rec.field_type)
        {
            case 'N':
                     field_off += field_rec.len_info.num_size.len;
                     break;
            default:
                     field_off += field_rec.len_info.char_len;
                     break;
        }
    }
                                       D(printf("%u",field_off);)
/*  7.  For each memo field record:                    */
                                       D(printf("\nUpdate DBF #"); )
    for (i = 0; i < num_memos; i++)
    {
/*      7.1  Update DBF record to point to             */
/*           new DBT record.                           */
        update_dbf(dbf_handle, dbf_recs[i], dbt_recs[i],
                   dbf_head.data_offset, field_off,
                   dbf_head.rec_size);
                                       D(printf(" %d ",i);)
    }
/*  WRAP UP & EXIT */
    close(dbf_handle);

    free((char *)dbf_recs);
    free((char *)dbt_recs);
                                       D(printf("\nMEM_PAK done. "); )
    _retl(TRUE);
}
/*******************************************************/
/* MOVE_MEMO()           BUFFERED
 *             Move entire memo field, handling those
 *             with more than one physical record, to
 *             a new location further down in the file.*/
/*******************************************************/
unsigned move_memo (int handle, unsigned from,
                    unsigned to )
{
    char *memo_rec;
    unsigned num_memo_recs, i;
    Boolean eom;

                               D(int npass = 0;)

                               D(printf("\nMoving MEMO HANDLE %d from ", handle);)
                               D(printf(" %u to %u.", from, to);)

    memo_rec = get_dbt_rec( handle, from);
    num_memo_recs = 1;

    eom = FALSE;

                               D(printf("\nPASS # / i =");)
    while (!eom)
    {
        /* Get size of last READ to terminating 0x1a = ASCII_EOF */
        i=0;
        while ( (i < MEMO_SIZE) && (memo_rec[i] != ASCII_EOF) )
              i++;
                               D(printf("| %d / %d |", npass++, i);)
                               D(if(!(npass % 4)) printf("\n            "); )

        if ( i == MEMO_SIZE)
        {
            put_dbt_rec(handle, to + num_memo_recs - 1, memo_rec);

            num_memo_recs = num_memo_recs + 1;

            memo_rec = get_dbt_rec(handle, from + num_memo_recs - 1);
        }
        else
        {
            eom = TRUE;
        }
    }
    put_dbt_rec (handle, to + num_memo_recs - 1, memo_rec);

    return(num_memo_recs);
}
/*******************************************************
 *
 * Set dbf field at record dbf_rec to point to new dbt_rec.
 *
 *******************************************************/

void update_dbf (int handle, unsigned dbf_rec, unsigned dbt_rec,
            unsigned data_off, unsigned field_off,
            unsigned rec_size)
{
    unsigned i, i1;
    char num_str[6];
    char num;
                            D(printf("\nUPDATE_DBF:  Convert %u to ",dbt_rec);)

    lseek (handle, (long) (data_off + field_off +
                          (dbf_rec - 1L) * rec_size), SEEK_SET);

    for (i=0; i<5; i++)
    {
        write(handle," ",1);
    }
    /* convert number to ASCII digit string */
    i = 10000;
    for ( i1 = 0; i1 < 5; i1++)
    {
        num = dbt_rec / i;
        if (num > 0)
        {
            dbt_rec = dbt_rec - num*i;
            num_str[i1] = num + '0';
        }
        else
        {
            num_str[i1] = '0';
        }
        i=i/10;
    }
    write(handle, num_str,5);
                               D(num_str[5] = '\0' ;      )
                               D(printf("%s",num_str);    )
}
/* EOF:  CMEMOPAK.C */
